@@ -36,6 +36,22 @@
 
 #define LINE_LEN 16
 
+static void x86emu_rdmsr(x86emu_t *emu)
+{
+  unsigned u = emu->x86.R_ECX;
+
+  emu->x86.R_EDX = emu->x86.msr[u] >> 32;
+  emu->x86.R_EAX = emu->x86.msr[u];
+  emu->x86.msr_perm[u] |= X86EMU_ACC_R;
+}
+
+static void x86emu_wrmsr(x86emu_t *emu)
+{
+  unsigned u = emu->x86.R_ECX;
+
+  emu->x86.msr[u] = ((u64) emu->x86.R_EDX << 32) + emu->x86.R_EAX;
+  emu->x86.msr_perm[u] |= X86EMU_ACC_W;
+}
 
 API_SYM x86emu_t *x86emu_new(unsigned def_mem_perm, unsigned def_io_perm)
 {
@@ -50,6 +66,8 @@ API_SYM x86emu_t *x86emu_new(unsigned def_mem_perm, unsigned def_io_perm)
   if(def_io_perm) x86emu_set_io_perm(emu, 0, X86EMU_IO_PORTS - 1, def_io_perm);
 
   x86emu_set_memio_handler(emu, vm_memio);
+  x86emu_set_rdmsr_handler(emu, x86emu_rdmsr);
+  x86emu_set_wrmsr_handler(emu, x86emu_wrmsr);
 
   x86emu_reset(emu);
 
@@ -179,6 +197,41 @@ API_SYM x86emu_code_handler_t x86emu_set_code_handler(x86emu_t *emu, x86emu_code
   return old;
 }
 
+API_SYM x86emu_cpuid_handler_t x86emu_set_cpuid_handler(x86emu_t *emu, x86emu_cpuid_handler_t handler)
+{
+  x86emu_cpuid_handler_t old = NULL;
+
+  if(emu) {
+    old = emu->cpuid;
+    emu->cpuid = handler;
+  }
+
+  return old;
+}
+
+API_SYM x86emu_wrmsr_handler_t x86emu_set_wrmsr_handler(x86emu_t *emu, x86emu_wrmsr_handler_t handler)
+{
+  x86emu_wrmsr_handler_t old = NULL;
+
+  if(emu) {
+    old = emu->wrmsr;
+    emu->wrmsr = handler;
+  }
+
+  return old;
+}
+
+API_SYM x86emu_rdmsr_handler_t x86emu_set_rdmsr_handler(x86emu_t *emu, x86emu_rdmsr_handler_t handler)
+{
+  x86emu_rdmsr_handler_t old = NULL;
+
+  if(emu) {
+    old = emu->rdmsr;
+    emu->rdmsr = handler;
+  }
+
+  return old;
+}
 
 API_SYM unsigned x86emu_read_byte(x86emu_t *emu, unsigned addr)
 {
