@@ -114,8 +114,9 @@ u16     xor_word (x86emu_t *emu, u16 d, u16 s);
 u32     xor_long (x86emu_t *emu, u32 d, u32 s);
 void    imul_byte (x86emu_t *emu, u8 s);
 void    imul_word (x86emu_t *emu, u16 s);
+int 	imul_word_direct(u16 *res_lo, u16* res_hi, u16 d, u16 s);
 void    imul_long (x86emu_t *emu, u32 s);
-void 	imul_long_direct(u32 *res_lo, u32* res_hi, u32 d, u32 s);
+int 	imul_long_direct(u32 *res_lo, u32* res_hi, u32 d, u32 s);
 void    mul_byte (x86emu_t *emu, u8 s);
 void    mul_word (x86emu_t *emu, u16 s);
 void    mul_long (x86emu_t *emu, u32 s);
