@@ -280,6 +280,209 @@ static void x86emuOp2_wbinvd(x86emu_t *emu, u8 op2)
   OP_DECODE("wbinvd");
 }
 
+static void x86emuOp2_sse_enabled_check(x86emu_t *emu)
+{
+  if (emu->x86.R_CR0 & CR0_EM || !(emu->x86.R_CR0 & CR0_MP))
+    INTR_RAISE_UD(emu);
+  if (!(emu->x86.R_CR4 & CR4_OSFXSR))
+    INTR_RAISE_UD(emu);
+}
+
+/****************************************************************************
+REMARKS:
+Handles opcode 0x0f,0x10-0x11 0xf,0x28-0x29 0xf,0x2b-0x2c
+****************************************************************************/
+static void x86emuOp2_SSEmovops(x86emu_t *emu, u8 op2)
+{
+  int mod, rl, rh;
+  I128_reg_t *src, *dst, tmp;
+  u32 addr = 0;
+
+  switch(op2 & 0xfe) {
+    case 0x10:
+      OP_DECODE("movups ");
+      break;
+    case 0x28:
+      OP_DECODE("movaps ");
+      break;
+    case 0x2b:
+      OP_DECODE("movntps ");
+      break;
+  }
+
+  x86emuOp2_sse_enabled_check(emu);
+
+  fetch_decode_modrm(emu, &mod, &rh, &rl);
+
+  if(mod == 3) {
+    dst = decode_rm_sse_register(emu, rh);
+    OP_DECODE(",");
+    src = decode_rm_sse_register(emu, rl);
+    memcpy(dst, src, sizeof(*dst));
+  }
+  else {
+    if(!(op2 & 1)) {
+      dst = decode_rm_sse_register(emu, rh);
+      OP_DECODE(",");
+      addr = decode_rm_address(emu, mod, rl);
+      tmp = fetch_data_qlong(emu, addr);
+      /* check memory alignment if not movups */
+      if (((op2 & 0xfe) != 0x10) && (addr & 0xf)) {
+        INTR_RAISE_GP(emu, 0);
+        return;
+      }
+      memcpy(dst, &tmp, sizeof(*dst));
+    } else {
+      addr = decode_rm_address(emu, mod, rh);
+      OP_DECODE(",");
+      src = decode_rm_sse_register(emu, rl);
+      /* check memory alignment if not movups */
+      if(((op2 & 0xfe) != 0x10) && (addr & 0xf)) {
+        INTR_RAISE_GP(emu, 0);
+        return;
+      }
+      memcpy(&tmp, src, sizeof(tmp));
+      store_data_qlong(emu, addr, tmp);
+    }
+  }
+}
+
+/****************************************************************************
+REMARKS:
+Handles opcode 0x0f,0x12-0x13 0x0f,0x16-0x17
+****************************************************************************/
+static void x86emuOp2_SSEmovpackedops(x86emu_t *emu, u8 op2)
+{
+  int mod, rl, rh, i;
+  I128_reg_t *src, *dst;
+  u32 addr;
+
+  fetch_decode_modrm(emu, &mod, &rh, &rl);
+
+  switch(op2) {
+    case 0x12:
+      if (mod == 3)
+        OP_DECODE("movhlps ");
+      else
+        OP_DECODE("movlps ");
+      break;
+    case 0x13:
+      OP_DECODE("movlps ");
+      break;
+    case 0x16:
+      if (mod == 3)
+        OP_DECODE("movhlps ");
+      else
+        OP_DECODE("movhps ");
+      break;
+    case 0x17:
+      OP_DECODE("movhps ");
+      break;
+    default:
+      INTR_RAISE_UD(emu);
+      break;
+  }
+
+  x86emuOp2_sse_enabled_check(emu);
+
+  if(mod == 3) {
+    dst = decode_rm_sse_register(emu, rh);
+    OP_DECODE(",");
+    src = decode_rm_sse_register(emu, rl);
+    for(i = 0; i < FP_SP_SIZE * 2; i++) {
+      if (op2 == 0x12) {
+        dst->reg[i] = src->reg[i + FP_SP_SIZE * 2];
+      } else if (op2 == 0x16) {
+        dst->reg[i+ FP_SP_SIZE * 2] = src->reg[i];
+      }
+    }
+  }else{
+    if (op2 & 1) {
+      addr = decode_rm_address(emu, mod, rl);
+      OP_DECODE(",");
+      src = decode_rm_sse_register(emu, rh);
+
+      for(i = 0; i < FP_SP_SIZE * 2; i++) {
+        if (op2 == 0x13) {
+          store_data_byte(emu, addr + i, src->reg[i]);
+        } else if (op2 == 0x17) {
+          store_data_byte(emu, addr + i, src->reg[i+FP_SP_SIZE*2]);
+        }
+      }
+    } else {
+      dst = decode_rm_sse_register(emu, rh);
+      OP_DECODE(",");
+      addr = decode_rm_address(emu, mod, rl);
+
+      for(i = 0; i < FP_SP_SIZE * 2; i++) {
+        if (op2 == 0x12) {
+          dst->reg[i] = fetch_data_byte(emu, addr + i);
+        } else if (op2 == 0x16) {
+          dst->reg[i+FP_SP_SIZE*2] = fetch_data_byte(emu, addr + i);
+        }
+      }
+    }
+  }
+}
+
+
+/****************************************************************************
+REMARKS:
+Handles opcode 0x0f,0x14-0x15
+****************************************************************************/
+static void x86emuOp2_SSEpackops(x86emu_t *emu, u8 op2)
+{
+  int mod, rl, rh, off, i;
+  I128_reg_t *src, *dst, tmp, arith;
+  u32 addr;
+
+  switch(op2) {
+    case 0x14:
+      OP_DECODE("unpcklps ");
+      break;
+    case 0x15:
+      OP_DECODE("unpckhps ");
+      break;
+    default:
+      INTR_RAISE_UD(emu);
+      break;
+  }
+
+  x86emuOp2_sse_enabled_check(emu);
+
+  fetch_decode_modrm(emu, &mod, &rh, &rl);
+
+  if(mod == 3) {
+    dst = decode_rm_sse_register(emu, rh);
+    OP_DECODE(",");
+    src = decode_rm_sse_register(emu, rl);
+  } else {
+    dst = decode_rm_sse_register(emu, rh);
+    OP_DECODE(",");
+    addr = decode_rm_address(emu, mod, rl);
+    tmp = fetch_data_qlong(emu, addr);
+    src = &tmp;
+  }
+
+  if(op2 == 0x14) {
+    off = 0;
+  } else {
+    off = 2;
+  }
+
+  arith.reg[0] = dst->reg[off + 0];
+  arith.reg[1] = src->reg[off + 0];
+  arith.reg[2] = dst->reg[off + 1];
+  arith.reg[3] = src->reg[off + 1];
+  arith.reg[4] = dst->reg[off + 4];
+  arith.reg[5] = src->reg[off + 4];
+  arith.reg[6] = dst->reg[off + 5];
+  arith.reg[7] = src->reg[off + 5];
+
+  for(i = 0; i < FP_SP_SIZE; i++) {
+    dst->reg[i] = arith.reg[i];
+  }
+}
 
 /****************************************************************************
 REMARKS:
@@ -389,8 +592,9 @@ static void x86emuOp2_wrmsr(x86emu_t *emu, u8 op2)
     INTR_RAISE_UD(emu);
   }
   else {
-    emu->x86.msr[u] = ((u64) emu->x86.R_EDX << 32) + emu->x86.R_EAX;
-    emu->x86.msr_perm[u] |= X86EMU_ACC_W;
+    if(emu->wrmsr) {
+      emu->wrmsr(emu);
+    }
   }
 }
 
@@ -421,14 +625,13 @@ static void x86emuOp2_rdmsr(x86emu_t *emu, u8 op2)
   OP_DECODE("rdmsr");
 
   u = emu->x86.R_ECX;
-
   if(u >= X86EMU_MSRS) {
     INTR_RAISE_UD(emu);
   }
   else {
-    emu->x86.R_EDX = emu->x86.msr[u] >> 32;
-    emu->x86.R_EAX = emu->x86.msr[u];
-    emu->x86.msr_perm[u] |= X86EMU_ACC_R;
+    if(emu->rdmsr) {
+      emu->rdmsr(emu);
+    }
   }
 }
 
@@ -595,6 +798,68 @@ static void x86emuOp2_conditional_move(x86emu_t *emu, u8 op2)
   }
 }
 
+/****************************************************************************
+REMARKS:
+Handles opcode 0x0f,0x54-57
+****************************************************************************/
+static void x86emuOp2_SSElogicalops(x86emu_t *emu, u8 op2)
+{
+  int mod, rl, rh, i;
+  I128_reg_t *src, *dst, tmp;
+  u32 addr;
+
+  switch(op2) {
+    case 0x54:
+      OP_DECODE("andps ");
+      break;
+    case 0x55:
+      OP_DECODE("andnps ");
+      break;
+    case 0x56:
+      OP_DECODE("orps ");
+      break;
+    case 0x57:
+      OP_DECODE("xorps ");
+      break;
+    default:
+      INTR_RAISE_UD(emu);
+      break;
+  }
+
+  x86emuOp2_sse_enabled_check(emu);
+
+  fetch_decode_modrm(emu, &mod, &rh, &rl);
+
+  if(mod == 3) {
+    dst = decode_rm_sse_register(emu, rh);
+    OP_DECODE(",");
+    src = decode_rm_sse_register(emu, rl);
+  }
+  else {
+    dst = decode_rm_sse_register(emu, rh);
+    OP_DECODE(",");
+    addr = decode_rm_address(emu, mod, rl);
+    tmp = fetch_data_qlong(emu, addr);
+    src = &tmp;
+  }
+  for(i = 0; i < sizeof(dst->reg); i++) {
+    switch(op2) {
+      case 0x54:
+        dst->reg[i] &= src->reg[i];
+        break;
+      case 0x55:
+        dst->reg[i] &= ~src->reg[i];
+        break;
+      case 0x56:
+        dst->reg[i] |= src->reg[i];
+        break;
+      case 0x57:
+        dst->reg[i] ^= src->reg[i];
+        break;
+    }
+  }
+}
+
 /****************************************************************************
 REMARKS:
 Handles opcode 0x0f,0x80-0x8F
@@ -679,6 +944,21 @@ static void x86emuOp2_pop_FS(x86emu_t *emu, u8 op2)
   x86emu_set_seg_register(emu, emu->x86.R_FS_SEL, MODE_DATA32 ? pop_long(emu) : pop_word(emu));
 }
 
+/****************************************************************************
+REMARKS:
+Handles opcode 0x0f,0xa2
+****************************************************************************/
+static void x86emuOp2_cpuid(x86emu_t *emu, u8 op2)
+{
+  OP_DECODE("cpuid ");
+
+  if(emu->cpuid) {
+    emu->cpuid(emu);
+  }
+  else {
+    INTR_RAISE_UD(emu);
+  }
+}
 
 /****************************************************************************
 REMARKS:
@@ -1043,8 +1323,8 @@ Handles opcode 0x0f,0xaf
 static void x86emuOp2_imul_R_RM(x86emu_t *emu, u8 op2)
 {
   int mod, rl, rh;
-  u32 *src32, *dst32, val, addr, res_lo, res_hi;
-  u16 *src16, *dst16;
+  u32 *src32, *dst32, val32, res32_lo, res32_hi, addr;
+  u16 *src16, *dst16, val16, res16_lo, res16_hi;
   
   OP_DECODE("imul ");
   fetch_decode_modrm(emu, &mod, &rh, &rl);
@@ -1054,8 +1334,7 @@ static void x86emuOp2_imul_R_RM(x86emu_t *emu, u8 op2)
       dst32 = decode_rm_long_register(emu, rh);
       OP_DECODE(",");
       src32 = decode_rm_long_register(emu, rl);
-      imul_long_direct(&res_lo, &res_hi, *dst32, *src32);
-      if(res_hi != 0) {
+      if(imul_long_direct(&res32_lo, &res32_hi, *dst32, *src32)) {
         SET_FLAG(F_CF);
         SET_FLAG(F_OF);
       }
@@ -1063,14 +1342,15 @@ static void x86emuOp2_imul_R_RM(x86emu_t *emu, u8 op2)
         CLEAR_FLAG(F_CF);
         CLEAR_FLAG(F_OF);
       }
-      *dst32= res_lo;
+      *dst32= res32_lo;
+
+      SET_FLAGS_FOR_MUL(res32_hi | res32_lo, res32_lo, 32);
     }
     else {
       dst16 = decode_rm_word_register(emu, rh);
       OP_DECODE(",");
       src16 = decode_rm_word_register(emu, rl);
-      res_lo = (s32) ((s16) *dst16 * (s16) *src16);
-      if(res_lo > 0xffff) {
+      if(imul_word_direct(&res16_lo, &res16_hi, *dst16, *src16)) {
         SET_FLAG(F_CF);
         SET_FLAG(F_OF);
       }
@@ -1078,7 +1358,9 @@ static void x86emuOp2_imul_R_RM(x86emu_t *emu, u8 op2)
         CLEAR_FLAG(F_CF);
         CLEAR_FLAG(F_OF);
       }
-      *dst16 = res_lo;
+      *dst16 = res16_lo;
+
+      SET_FLAGS_FOR_MUL(res16_hi | res16_lo, res16_lo, 16);
     }
   }
   else {
@@ -1086,9 +1368,8 @@ static void x86emuOp2_imul_R_RM(x86emu_t *emu, u8 op2)
       dst32 = decode_rm_long_register(emu, rh);
       OP_DECODE(",");
       addr = decode_rm_address(emu, mod, rl);
-      val = fetch_data_long(emu, addr);
-      imul_long_direct(&res_lo, &res_hi, *dst32, val);
-      if(res_hi != 0) {
+      val32 = fetch_data_long(emu, addr);
+      if(imul_long_direct(&res32_lo, &res32_hi, *dst32, val32)) {
         SET_FLAG(F_CF);
         SET_FLAG(F_OF);
       }
@@ -1096,15 +1377,16 @@ static void x86emuOp2_imul_R_RM(x86emu_t *emu, u8 op2)
         CLEAR_FLAG(F_CF);
         CLEAR_FLAG(F_OF);
       }
-      *dst32 = res_lo;
+      *dst32 = res32_lo;
+
+      SET_FLAGS_FOR_MUL(res32_hi | res32_lo, res32_lo, 32);
     }
     else {
       dst16 = decode_rm_word_register(emu, rh);
       OP_DECODE(",");
       addr = decode_rm_address(emu, mod, rl);
-      val = fetch_data_word(emu, addr);
-      res_lo = (s32) ((s16) *dst16 * (s16) val);
-      if(res_lo > 0xffff) {
+      val16 = fetch_data_word(emu, addr);
+      if(imul_word_direct(&res16_lo, &res16_hi, *dst16, val16)) {
         SET_FLAG(F_CF);
         SET_FLAG(F_OF);
       }
@@ -1112,7 +1394,9 @@ static void x86emuOp2_imul_R_RM(x86emu_t *emu, u8 op2)
         CLEAR_FLAG(F_CF);
         CLEAR_FLAG(F_OF);
       }
-      *dst16 = res_lo;
+      *dst16 = res16_lo;
+
+      SET_FLAGS_FOR_MUL(res16_hi | res16_lo, res16_lo, 16);
     }
   }
 }
@@ -1781,14 +2065,14 @@ void (*x86emu_optab2[256])(x86emu_t *emu, u8) =
   /*  0x0e */ x86emuOp2_illegal_op,
   /*  0x0f */ x86emuOp2_illegal_op,
 
-  /*  0x10 */ x86emuOp2_illegal_op,
-  /*  0x11 */ x86emuOp2_illegal_op,
-  /*  0x12 */ x86emuOp2_illegal_op,
-  /*  0x13 */ x86emuOp2_illegal_op,
-  /*  0x14 */ x86emuOp2_illegal_op,
-  /*  0x15 */ x86emuOp2_illegal_op,
-  /*  0x16 */ x86emuOp2_illegal_op,
-  /*  0x17 */ x86emuOp2_illegal_op,
+  /*  0x10 */ x86emuOp2_SSEmovops,
+  /*  0x11 */ x86emuOp2_SSEmovops,
+  /*  0x12 */ x86emuOp2_SSEmovpackedops,
+  /*  0x13 */ x86emuOp2_SSEmovpackedops,
+  /*  0x14 */ x86emuOp2_SSEpackops,
+  /*  0x15 */ x86emuOp2_SSEpackops,
+  /*  0x16 */ x86emuOp2_SSEmovpackedops,
+  /*  0x17 */ x86emuOp2_SSEmovpackedops,
   /*  0x18 */ x86emuOp2_illegal_op,
   /*  0x19 */ x86emuOp2_illegal_op,
   /*  0x1a */ x86emuOp2_illegal_op,
@@ -1806,11 +2090,11 @@ void (*x86emu_optab2[256])(x86emu_t *emu, u8) =
   /*  0x25 */ x86emuOp2_illegal_op,
   /*  0x26 */ x86emuOp2_illegal_op,  /* mov treg,reg32 (ring 0 PM) */
   /*  0x27 */ x86emuOp2_illegal_op,
-  /*  0x28 */ x86emuOp2_illegal_op,
-  /*  0x29 */ x86emuOp2_illegal_op,
+  /*  0x28 */ x86emuOp2_SSEmovops,
+  /*  0x29 */ x86emuOp2_SSEmovops,
   /*  0x2a */ x86emuOp2_illegal_op,
-  /*  0x2b */ x86emuOp2_illegal_op,
-  /*  0x2c */ x86emuOp2_illegal_op,
+  /*  0x2b */ x86emuOp2_SSEmovops,
+  /*  0x2c */ x86emuOp2_SSEmovops,
   /*  0x2d */ x86emuOp2_illegal_op,
   /*  0x2e */ x86emuOp2_illegal_op,
   /*  0x2f */ x86emuOp2_illegal_op,
@@ -1853,10 +2137,10 @@ void (*x86emu_optab2[256])(x86emu_t *emu, u8) =
   /*  0x51 */ x86emuOp2_illegal_op,
   /*  0x52 */ x86emuOp2_illegal_op,
   /*  0x53 */ x86emuOp2_illegal_op,
-  /*  0x54 */ x86emuOp2_illegal_op,
-  /*  0x55 */ x86emuOp2_illegal_op,
-  /*  0x56 */ x86emuOp2_illegal_op,
-  /*  0x57 */ x86emuOp2_illegal_op,
+  /*  0x54 */ x86emuOp2_SSElogicalops,
+  /*  0x55 */ x86emuOp2_SSElogicalops,
+  /*  0x56 */ x86emuOp2_SSElogicalops,
+  /*  0x57 */ x86emuOp2_SSElogicalops,
   /*  0x58 */ x86emuOp2_illegal_op,
   /*  0x59 */ x86emuOp2_illegal_op,
   /*  0x5a */ x86emuOp2_illegal_op,
@@ -1936,7 +2220,7 @@ void (*x86emu_optab2[256])(x86emu_t *emu, u8) =
 
   /*  0xa0 */ x86emuOp2_push_FS,
   /*  0xa1 */ x86emuOp2_pop_FS,
-  /*  0xa2 */ x86emuOp2_illegal_op,
+  /*  0xa2 */ x86emuOp2_cpuid,
   /*  0xa3 */ x86emuOp2_bt_R,
   /*  0xa4 */ x86emuOp2_shld_IMM,
   /*  0xa5 */ x86emuOp2_shld_CL,
